﻿#region Using Statements
using System;
using System.Collections.Generic;
using System.IO;
using Microsoft.Xna.Framework;
using Microsoft.Xna.Framework.Audio;
using Microsoft.Xna.Framework.Content;
using Microsoft.Xna.Framework.Graphics;
using Microsoft.Xna.Framework.Input;
using Microsoft.Xna.Framework.Storage;
#endregion

namespace reimers_tutorial13_series1
{
    /// <summary>
    /// This is the main type for your game
    /// </summary>
    public class Game1 : Microsoft.Xna.Framework.Game
    {
        public struct VertexPositionNormalColored
        {
            public Vector3 Position;
            public Color Color;
            public Vector3 Normal;

            public static int SizeInBytes = 7 * 4;

            public static VertexElement[] VertexElements = new VertexElement[]
            {
                new VertexElement(0, 0, VertexElementFormat.Vector3, VertexElementMethod.Default, VertexElementUsage.Position, 0),
                new VertexElement(0, sizeof(float) * 3, VertexElementFormat.Color, VertexElementMethod.Default, VertexElementUsage.Color, 0),
                new VertexElement(0, sizeof(float) * 4, VertexElementFormat.Vector3, VertexElementMethod.Default, VertexElementUsage.Normal, 0)
            };
        }

        GraphicsDeviceManager graphics;
        ContentManager content;

        GraphicsDevice device;

        Effect effect;

        VertexPositionNormalColored[] vertices;

        VertexBuffer vb;
        IndexBuffer ib;

        float angle = 0.0f;

        int WIDTH = 64;
        int HEIGHT = 64;

        int[,] heightdata;

        int minimumheight = 255;
        int maximumheight = 0;

        public Game1()
        {
            graphics = new GraphicsDeviceManager(this);
            content = new ContentManager(Services);
        }


        /// <summary>
        /// Allows the game to perform any initialization it needs to before starting to run.
        /// This is where it can query for any required services and load any non-graphic
        /// related content.  Calling base.Initialize will enumerate through any components
        /// and initialize them as well.
        /// </summary>
        protected override void Initialize()
        {
            // TODO: Add your initialization logic here

            base.Initialize();

            LoadHeightData();

            SetupXNADevice();

            SetupVertices();
            SetupIndices();

            SetupCamera();
        }

        private void SetupXNADevice()
        {
            device = graphics.GraphicsDevice;

            graphics.PreferredBackBufferWidth = 500;
            graphics.PreferredBackBufferHeight = 500;

            graphics.IsFullScreen = false;

            graphics.ApplyChanges();

            Window.Title = "riemer`s XNA tutorial 13 - series 1";

            CompiledEffect compiledEffect = Effect.CompileEffectFromFile("@/../../../../effects.fx", null, null, CompilerOptions.None, TargetPlatform.Windows);
            effect = new Effect(graphics.GraphicsDevice, compiledEffect.GetEffectCode(), CompilerOptions.None, null);
        }

        private void SetupCamera()
        {
            Matrix viewMatrix = Matrix.CreateLookAt(new Vector3(80.0f, 0.0f, 160.0f), new Vector3(-20.0f, 0.0f, 0.0f), new Vector3(0.0f, 0.0f, 1.0f));
            Matrix projectionMatrix = Matrix.CreatePerspectiveFieldOfView(MathHelper.PiOver4, Window.ClientBounds.Width / Window.ClientBounds.Height, 1.0f, 250.0f);

            effect.Parameters["xView"].SetValue(viewMatrix);
            effect.Parameters["xProjection"].SetValue(projectionMatrix);
            effect.Parameters["xWorld"].SetValue(Matrix.Identity);
        }

        private void LoadHeightData()
        {
            FileStream fs = new FileStream("../../../heightmap.bmp", FileMode.Open, FileAccess.Read);
            BinaryReader br = new BinaryReader(fs);

            int offset;

            br.BaseStream.Seek(10, SeekOrigin.Current);
            offset = (int)br.ReadUInt32();

            br.BaseStream.Seek(4, SeekOrigin.Current);
            WIDTH = (int)br.ReadUInt32();
            HEIGHT = (int)br.ReadUInt32();

            br.BaseStream.Seek(offset - 26, SeekOrigin.Current);
            heightdata = new int[WIDTH, HEIGHT];
            for (int y = 0; y < HEIGHT; y++)
            {
                for (int x = 0; x < WIDTH; x++)
                {
                    int height = ((int)(br.ReadByte()) + (int)(br.ReadByte()) + (int)(br.ReadByte())) / 8;
                    heightdata[WIDTH - 1 - x, HEIGHT - 1 - y] = height;

                    if (height < minimumheight)
                        minimumheight = height;
                    if (height > maximumheight)
                        maximumheight = height;
                }
            }

            br.Close();
        }

        private void SetupVertices()
        {
            vertices = new VertexPositionNormalColored[WIDTH * HEIGHT];

            for (int x = 0; x < WIDTH; x++)
            {
                for (int y = 0; y < HEIGHT; y++)
                {
                    vertices[x + y * WIDTH].Position = new Vector3(x, y, heightdata[x,y]);
                    vertices[x + y * WIDTH].Normal = new Vector3(0.0f, 0.0f, 1.0f);

                    if (heightdata[x, y] < minimumheight + (maximumheight - minimumheight) * 1 / 4)
                        vertices[x + y * WIDTH].Color = Color.Blue;
                    else if (heightdata[x, y] < minimumheight + (maximumheight - minimumheight) * 2 / 4)
                        vertices[x + y * WIDTH].Color = Color.Green;
                    else if (heightdata[x, y] < minimumheight + (maximumheight - minimumheight) * 3 / 4)
                        vertices[x + y * WIDTH].Color = Color.Brown;
                    else
                        vertices[x + y * WIDTH].Color = Color.White;
                }
            }

            for (int x = 1; x < WIDTH - 1; x++)
            {
                for (int y = 1; y < HEIGHT - 1; y++)
                {
                    Vector3 normX = new Vector3((vertices[x - 1 + y * WIDTH].Position.Z - vertices[x + 1 + y * WIDTH].Position.Z) / 2, 0, 1);
                    Vector3 normY = new Vector3(0, (vertices[x + (y - 1) * WIDTH].Position.Z - vertices[x + (y + 1) * WIDTH].Position.Z) / 2, 1);
                    vertices[x + y * WIDTH].Normal = normX + normY;
                    vertices[x + y * WIDTH].Normal.Normalize();
                }
            }

            vb = new VertexBuffer(device, VertexPositionNormalColored.SizeInBytes * WIDTH * HEIGHT, ResourceUsage.WriteOnly, ResourceManagementMode.Automatic);
            vb.SetData(vertices);
        }

        private void SetupIndices()
        {
            int[] indices = new int[(WIDTH - 1) * (HEIGHT - 1) * 3 * 2];

            for (int x = 0; x < WIDTH - 1; x++)
            {
                for (int y = 0; y < HEIGHT - 1; y++)
                {
                    indices[(x + y * (WIDTH - 1)) * 3 * 2 + 0] = (x + 1) + (y + 1) * WIDTH;
                    indices[(x + y * (WIDTH - 1)) * 3 * 2 + 1] = (x + 1) + (y + 0) * WIDTH;
                    indices[(x + y * (WIDTH - 1)) * 3 * 2 + 2] = (x + 0) + (y + 0) * WIDTH;

                    indices[(x + y * (WIDTH - 1)) * 3 * 2 + 3] = (x + 1) + (y + 1) * WIDTH;
                    indices[(x + y * (WIDTH - 1)) * 3 * 2 + 4] = (x + 0) + (y + 0) * WIDTH;
                    indices[(x + y * (WIDTH - 1)) * 3 * 2 + 5] = (x + 0) + (y + 1) * WIDTH;
                }
            }

            ib = new IndexBuffer(device, typeof(int), (WIDTH - 1) * (HEIGHT -1) * 3 * 2, ResourceUsage.WriteOnly, ResourceManagementMode.Automatic);
            ib.SetData(indices);
        }

        /// <summary>
        /// Load your graphics content.  If loadAllContent is true, you should
        /// load content from both ResourceManagementMode pools.  Otherwise, just
        /// load ResourceManagementMode.Manual content.
        /// </summary>
        /// <param name="loadAllContent">Which type of content to load.</param>
        protected override void LoadGraphicsContent(bool loadAllContent)
        {
            if (loadAllContent)
            {
                // TODO: Load any ResourceManagementMode.Automatic content
            }

            // TODO: Load any ResourceManagementMode.Manual content
        }


        /// <summary>
        /// Unload your graphics content.  If unloadAllContent is true, you should
        /// unload content from both ResourceManagementMode pools.  Otherwise, just
        /// unload ResourceManagementMode.Manual content.  Manual content will get
        /// Disposed by the GraphicsDevice during a Reset.
        /// </summary>
        /// <param name="unloadAllContent">Which type of content to unload.</param>
        protected override void UnloadGraphicsContent(bool unloadAllContent)
        {
            if (unloadAllContent)
            {
                // TODO: Unload any ResourceManagementMode.Automatic content
                content.Unload();
            }

            // TODO: Unload any ResourceManagementMode.Manual content
        }


        /// <summary>
        /// Allows the game to run logic such as updating the world,
        /// checking for collisions, gathering input and playing audio.
        /// </summary>
        /// <param name="gameTime">Provides a snapshot of timing values.</param>
        protected override void Update(GameTime gameTime)
        {
            // Allows the game to exit
            if (GamePad.GetState(PlayerIndex.One).Buttons.Back == ButtonState.Pressed)
                this.Exit();

            // TODO: Add your update logic here

            base.Update(gameTime);

            ProcessKeyboard();
        }

        private void ProcessKeyboard()
        {
            KeyboardState keys = Keyboard.GetState();

            if (keys.IsKeyDown(Keys.PageUp))
            {
                angle += 0.05f;
            }
            if (keys.IsKeyDown(Keys.PageDown))
            {
                angle -= 0.05f;
            }
        }


        /// <summary>
        /// This is called when the game should draw itself.
        /// </summary>
        /// <param name="gameTime">Provides a snapshot of timing values.</param>
        protected override void Draw(GameTime gameTime)
        {
            device.RenderState.CullMode = CullMode.None;

            device.Clear(ClearOptions.Target | ClearOptions.DepthBuffer, Color.Black, 1.0f, 0);

            // TODO: Add your drawing code here

            effect.CurrentTechnique = effect.Techniques["Colored"];

            Matrix worldMatrix = Matrix.CreateTranslation(-HEIGHT / 2, -WIDTH / 2, 0) * Matrix.CreateRotationZ(angle);

            effect.Parameters["xWorld"].SetValue(worldMatrix);
            effect.Parameters["xEnableLighting"].SetValue(true);
            effect.Parameters["xLightDirection"].SetValue(new Vector3(0.5f, 0.5f, -1.0f));

            effect.Begin();

            foreach (EffectPass pass in effect.CurrentTechnique.Passes)
            {
                pass.Begin();

                device.Vertices[0].SetSource(vb, 0, VertexPositionNormalColored.SizeInBytes);
                device.Indices = ib;
                device.VertexDeclaration = new VertexDeclaration(device, VertexPositionNormalColored.VertexElements);
                device.DrawIndexedPrimitives(PrimitiveType.TriangleList, 0, 0, WIDTH * HEIGHT, 0, (WIDTH - 1) * (HEIGHT - 1) * 2);
                
                pass.End();
            }

            effect.End();

            base.Draw(gameTime);
        }
    }
}
